
#include <inttypes.h>
#include <avr/io.h>
#include "button.h"


uint8_t button_read(void) {
  static uint8_t prev_input, button_count, status;
  uint8_t input, dinput, state;

  // detect events
  input = BUTTON_PORT;
  dinput = prev_input ^ input;

  state = 0;

  // WARNING
  // The gcc optimization setting -O2 optimizes the 
  // (dinput & _BV(BUTTON_PIN)) AWAY, use -O0 for now.

  // a short press is detected upon the release of the button
  if (dinput & _BV(BUTTON_PIN)) {
    if (button_count >= NOISE_THRES) {
      if (button_count <= SHORT_PRESS)
	state = 1; // short press
    }
    button_count = 0;
    status &= ~_BV(handled_long_press);
  } else {
    // a long press is detected when the button is pressed for x time units

    if (!(input & _BV(BUTTON_PIN))) { // active low
      if (button_count < 255) {
	button_count++;
	if (button_count == LONG_PRESS && (!(status & _BV(handled_long_press)))) {
	  state = 2;
	  status |= _BV(handled_long_press);
	}
      } 
    } else {
      button_count = 0; // no need for counting the time when the button is released
    }
  }

  prev_input = input;
  return state;
}

