
#ifndef __EVENTQUEUE_H__
#define __EVENTQUEUE_H__

#include <avr/io.h>
#include <stdio.h>
#include <stdarg.h>

/* ---------------------- HARDWARE definitions: 
 */

/* we take statistics for this many pulses + 1 */
#define ROTBY	5
#define AGGRNUM	((1<<ROTBY)-1)
// #define ROTBY0	3

/* refuse periods smaller than 2000 usec (500 * 4usec)
 * this means 500 Hz, or 30000 events/min, 15000RPM 
 */ 
#define MINPER 500

/* before the injector opens, it fires appr. 1/3 of nominal rate
 * This is the correction in 4usec cycles :
 * eg. 10 means 40usec, appropriate where injector opens in 120 usec 
 */
#define INJ_OPEN_CORR	10

struct tachin_t {
	uint32_t	pwsum;
	uint32_t	pwsqrsum;
	uint32_t	persum;
	uint32_t	persqrsum;
	uint16_t	rise_0;
	uint16_t	rise_prev;
	uint8_t		edge_cnt;
	uint8_t		edge_bad;
};

/* updated when calc ready, used by display */
struct pwdisplay_t {
	uint16_t	pwsum;
	uint16_t	pwsqrsum;
	uint16_t	persqrsum;
	uint16_t	persum;
	uint16_t	rpm;
	uint8_t		bad;
	uint8_t		dirty;
};

void eventqueue_init(void);
extern struct tachin_t tachin;
extern struct pwdisplay_t pwdisplay;

void tachin_reset(void);
void pwdisplay_from_tachin(uint16_t now);
uint8_t pw2comm(uint8_t i);

#endif
