
#include <inttypes.h>
#include "heap.h"

/* WARNING: untested, experimental!
 * designed as the base of an eventqueue to efficiently find nearest event
 *
 * author: Marcell Gal
 * copyright: GPL
 */

/* 
 * IMPORTANT: userspace should cli/sei around all functions here  !!
 * an eventqueue will use 2 heaps, since 16 bit time overflows fast.
 * Imagine: now = 37000
 * 38000 is in the future, 6000 is also in the future, actually later.
 * 
 */

struct heap_t h1,h2;

__inline__ void heap_init(struct heap_t *h)
{
	h->heaplast = HEAPROOT - 1; /* it will be HEAPROOT after adding 1 element */
	h->heapunderrun = h->heapoverrun = 0;
}

/* --------------------
 * heap is 1 based (first index is HEAPROOT) 
 * change if going to base 0
 */
#if HEAPROOT==1
__inline__ heapindex_t parent(heapindex_t i)
{
	return i >> 1;
}

/* returns child with smaller index */
__inline__ heapindexx_t child0(heapindex_t i)
{
	return i + i ; 
	/* second child is next */
}
#else 
#warning check that HEAPROOT and subroutines match

#endif

/* returns child index that have smaller val */
__inline__ heapindexx_t childsmaller(struct heap_t *h, heapindex_t i)
{
	heapindex_t s, s1;

	s = child0(i);
	if (((s1=s+1) <= h->heaplast) && (heapval(h,s1) < heapval(h,s)))
		return s1;

/* Warning: returned index not checked against heaplast */
	return s;
}

/* TODO: this should get 2 event_t *   */
__inline__ void heapitemcopy(struct heap_t *h, heapindex_t from, heapindex_t to)
{
	heapkey(h, to) = heapkey(h, from);
	heapval(h, to) = heapval(h, from);
}

/* swaps 2 event_t in the heap */
__inline__ void heapswap(struct heap_t *h, heapindex_t i, heapindex_t ii)
{
	struct event_t dt ;
	dt.key = heapkey(h, ii); /* we could use index heapt[0] the, placefiller */
	dt.val = heapval(h, ii);
	heapitemcopy (h, i, ii);
	heapkey(h, i) = dt.key;
	heapval(h, i) = dt.val;
}

/*
 * inserts an event into the heap
 * From userspace ALWAYS call this with interrupts disabled,
 * since timer interrupt also manipulates events 
 */
void heap_insert(struct heap_t *h, uint16_t key, uint16_t val) 
{
	heapindex_t hole, hh;

	if (h->heaplast >= HEAPSIZE-1 ){
		h->heapoverrun ++;
		return;		/* drop without much warning */
	}

	hole = ++h->heaplast; /* still < HEAPSIZE */

	for ( ; hole > HEAPROOT && key < heapkey(h, hh=parent(hole)); hole = hh )
		heapitemcopy(h, hh, hole);

	heapkey(h, hole) = key;
	heapval(h, hole) = val;
}

/* deletes first (smallest key) element */
void heap_del1st(struct heap_t *p)
{
	heapindex_t h;
	heapindexx_t hh;	

	if(p->heaplast < HEAPROOT) { /* it was empty */
		p->heapunderrun++;
		return;
	}

	/* old ROOT element deleted. Do not care if copies onto self. */
	heapitemcopy(p, p->heaplast, HEAPROOT); 
	p->heaplast--;

	for ( h = HEAPROOT;  ; h=hh ){
		hh=childsmaller(p, h);
		if (hh > p->heaplast || heapkey(p, hh) >= heapkey(p, h)) 
			break;
		heapswap(p, hh, h);
	}
}
/*  heap_del1st and heap_insert in one EFFICIENT step.
 * often used from one event to fire next (eg. cylinder, 
 * or matching inj-stop event at inj-start
 */
void heap_replace1st(struct heap_t *p, uint16_t key, uint16_t val) 
{
	heapindex_t h;
	heapindexx_t hh;	
	heapkey(p, HEAPROOT) = key;
        heapval(p, HEAPROOT) = val;
	for ( h = HEAPROOT;  ; h=hh ){
		hh=childsmaller(p, h);
		if (hh > p->heaplast || heapkey(p, hh) >= heapkey(p, h)) 
			break;
		heapswap(p, hh, h);
	}
}
