#ifndef __HEAP_H__
#define __HEAP_H__

#define HEAPSIZE        32

#include <inttypes.h>

/* In the event struct 
 * we could have a function pointer (16bits) and a parameter for it, but
 * having a dispatcher is more space efficient. This way an event is
 * only 4 bytes. 
 * It's the user's responsibility to use the value * as he wants: eg. 
 * a dispathcer should call a function after 
 * peeking at the high bits of value, and pass the low bits as an argument.
 */

typedef	uint16_t key_t;
typedef	uint16_t val_t;

struct event_t {
        key_t key;
        val_t val; 
        /* could be a struct? The specific subroutine can have it's
        own structures if this is not enough */
};

#if     HEAPSIZE > 255
	typedef uint16_t heapindex_t ;
#else
	typedef uint8_t heapindex_t ;
#endif

/* some intermediate results need higher resolution sooner */
#if  HEAPSIZE > 126
	typedef uint16_t heapindexx_t ;
#else 
	typedef uint8_t heapindexx_t ;
# endif


/* this could be 0 with a little more overhead, but functions must
 * be changed to match
 */
#define HEAPROOT	1

/* Note that having separate arrays for key and val would
 * probably be more efficient. Use these wrappers to get val and key: */
#define heapkey(h,A)	h->heapt[A].key
#define heapval(h,A)	h->heapt[A].val

extern struct heap_t {
	struct event_t  heapt[HEAPSIZE];
	heapindex_t	heaplast;
	uint8_t	heapoverrun;	/* posterior check for serious errors */
	uint8_t	heapunderrun; 	/* should stay 0 */
} h1, h2;

void heap_init(struct heap_t *);

/* 
 * IMPORTANT: userspace should cli/sei around all functions here  !!
 */

void heap_insert(struct heap_t *, uint16_t key, uint16_t val);
void heap_del1st(struct heap_t *);
void heap_replace1st(struct heap_t *, uint16_t key, uint16_t val);

#endif
