
#include "keyboard.h"
#include <avr/io.h>
#include <inttypes.h>


/* returns char just pressed.
Do not call this too often, or too rare.
Appr. once in 0.1 sec is good.
It has some primitive anti-prelling:
we only return keypress event if previous 2
were both unpressed, and now pressed.

Returns 0 if no key is pressed.
Note: returns '0' if 0 is pressed.
*/

	/* small  key-history buffer for anti-prell */
	static uint16_t vec_prev0, vec_prev1;

	/* Should this go to a separate file??
	Note: index 0 is bit 0 in raw_vec.  */
/*	char keydefs[] = { '3','6','9','#',  '2','5','8','0',  '1','4','7','*' }; */
        uint8_t keydefs[] = { '#','9','6','3',  '0','8','5','2',  '*','7','4','1' };


uint8_t cooked_char(void)
{
	uint8_t i;
	uint16_t vec, ret;

	vec = read_raw_vec();

	/* just pressed if now cleared, but previous 2 was set */
	ret = vec | ~(vec_prev0 & vec_prev1);
	vec_prev0 = vec_prev1; 	/* rotate our small key-history buffer */
	vec_prev1 = vec;

	if ((ret & 0xFFF) == 0xFFF) return 0; /* nothing pressed now, fast in usual case */

	for (i=0; i<12; i++){
		if((ret & 1) == 0) return keydefs[i];
		ret >>= 1;
	}
	return 0;	// though never reached 
}

/* 
Returns a bitfield with 0 in the bits that are actually being pressed.

operation:
We pull each col LOW, one at a time.
If a key is pressed, the given LINE PORT PIN will be pulled low
against the LINE input pullup

This is normally only called from cooked_char() 
*/
uint16_t read_raw_vec(void)
{
	uint16_t vec;
	COL1DDR |= COL1MASK; 	/* output */
	COL1PORT &= ~COL1MASK;	/* pull LOW */
	vec = LINESPIN & LINESMASK;
	COL1DDR &= ~COL1MASK; 	/* input */

	COL2DDR |= COL2MASK; 	/* output */
	COL2PORT &= ~COL2MASK;	/* pull LOW */
	vec <<= 4;
	vec |= LINESPIN & LINESMASK;
	COL2DDR &= ~COL2MASK; 	/* input */

	COL3DDR |= COL3MASK; 	/* output */
	COL3PORT &= ~COL3MASK;	/* pull LOW */
	vec <<= 4;
	vec |= LINESPIN & LINESMASK;
	COL3DDR &= ~COL3MASK; 	/* input */

	return vec;
};

__inline__ void keyboard_init(void)
{
	/* lines inputs */
	LINESDDR &= ~LINESMASK ;
	/* activate pullups */
	LINESPORT |= LINESMASK;

	/* columns to input */
	COL1DDR &= ~COL1MASK;
	COL2DDR &= ~COL2MASK;
	COL3DDR &= ~COL3MASK;

	vec_prev0 = vec_prev1 = 0xFFFF; /* nothing was pressed */
};

