
// LCD is connected on PORTA high bits in 4bit mode, control on PORTD

#ifndef __LCD_H__
#define __LCD_H__

#include <avr/io.h>
#include <stdio.h>
#include <stdarg.h>

/* ---------------------- HARDWARE definitions: 
 * data lines are higher 4 bits(0xF0)
 * This is currently wired in lcd_cmd() and lcd_data()
 */

#ifdef JAPANESE_LCD
// our Japanese friends use a different charset
#define DEGREESYMBOL	0xDF
#else
#define DEGREESYMBOL	0xB0
#endif 

/* any symbol will do: */
#define OUT_OF_RANGE_CHAR 0xAD

// The lcd is connected to the upper 4 bits of porta (the same signals as the leds)
#define LCD_DATA        PORTA
#define LCD_DATAPIN     PINA
#define LCD_DATADDR     DDRA
#define LCD_DATAPIN     PINA

#ifdef LCD_2x16

#define LCD_LINES	2
#define LCD_COLS	16

#elif LCD_4x20

#define LCD_LINES	4
#define LCD_COLS	20

#else

#warning "No LCD size defined, check my_make"

#endif

#define LCD_2ND_LINE_ADDR 0x40
#define LCD_CHARS	(LCD_LINES*LCD_COLS)

/* Control lines are on the same port, but arbitrary bits */
#define LCD_CTRL        PORTD
#define LCD_CTRLDDR     DDRD 

/* falling edge triggers data latch (blue) */
#define ENABLE  _BV(2)
//#define ENABLE  _BV(1)

/* RS=0: instruction, RS=1: data (grey) */
#define RS      _BV(4)
//#define RS      _BV(2)

/* direction: 0: write, 1:read (purple) */
#ifdef MIK_LCD
#define DDIR   _BV(3)  // MegaSquirtAVR v1 doesn't have PD4 connected to a header :-(
// lcd backlight duty cycle is LCD_BACKLIGHT_DUTYCYCLE/256
// pwm frequency is 244 Hz
#define LCD_BACKLIGHT_DUTYCYCLE 25
#else
#define DDIR   _BV(3)
//#define DDIR   _BV(4)
#endif

/* ------------------- END of HARDWARE definitions */

typedef unsigned char byte;
typedef uint8_t lcdpos_t;

uint8_t commbuffer[10];

void lcd_delay(uint16_t p);
void toggle_E(void);
void lcd_cls(void);
void lcd_home(void);
void lcd_control(byte disonoff, byte curonoff, byte curblink);
void lcd_goto(byte mesto);
void lcd_putch(byte data);
void lcd_putstr(byte *data);
void printbin(byte x);
void printhex(byte i);
void print2(uint8_t x);
void print3(uint16_t x);
void print5(unsigned int x);
void print10(unsigned long x);
void lcd_init(void);
void lcd_waitmode(char m);
void lcd_testLOW(void);
void lcd_testHIGH(void);
void lcd_testrun(void);
uint8_t lcd_busy(void);
void lcd_newdata(void);
void lcd_refresh(void);
void int2BCD5(uint16_t bin, lcdpos_t index);

#define BCD_DIGIT4	2
#define BCD_DIGIT3	1
#define BCD_DIGIT2	0

void int2BCDn(uint16_t bin, lcdpos_t index, char tenpower8);

void lcd_go2index(lcdpos_t index);
void update_char(char c, lcdpos_t index);
void update_chars(char *c, lcdpos_t index);
uint8_t lcd_cache2comm(uint8_t i);
void busywait(void);
__inline__ void lcd_write(byte data);

int printf(const char *format, ...);
void lcd_putstr_P(const char *data);
__inline__ void lcd_setpos(lcdpos_t index);
__inline__ lcdpos_t lcd_getpos(void);

extern lcdpos_t lcd_next_pos;                                                            

#endif
