/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen, mik@caffrey.dk]          *
***********************************************************
*  -- History --                                          *
*    release 0.1 - 06-03-2003                             *
*            http://caffrey.dk/megasquirt                 *
*    initial Motorola version by Bowling & Grippo (2002)  *
*            http://www.bgsoflex.com/megasquirt.html      *
\*********************************************************/


#include "megasquirt.h"
#include "global.h"
#include "comm.h"
#include "fuelcalc.h"
#include "helpers.h"
#include "storage.h"
#include "adc.h"
#include "actuators.h"
#include "tables.h"
#include "ve.h"
#include "iac.h"

#include "eventqueue.h"

#ifdef  KEYBOARD
#include "keyboard.h"
#endif

#ifdef  LCD
#include "lcd.h"
#endif

#ifdef BUTTON
#include "button.h"
#endif

#include <avr/io.h>
#include <inttypes.h>
#include <avr/wdt.h>
#include <stdlib.h>
#include <avr/signal.h>
#include <avr/interrupt.h>
#include <avr/sfr_defs.h>
#include <avr/pgmspace.h>


/*
TODO:
  Check reset source at power-up
  Watchdog timer 
*/

#ifdef LCD
extern lcdpos_t lcd_next_pos;
#endif

extern int16_t idle_integral;
extern int16_t idle_error;
extern int16_t idle_rpmdot;
extern int16_t pos;



extern uint8_t sensors[];
extern volatile struct squirt_t inj_port1, inj_port2;
extern volatile struct engine_t engine;
extern volatile struct config_t config;
extern volatile struct time_t rtc;
extern struct corr_t corr;
extern volatile struct iac_t iac;
extern uint8_t idle_rpm;

extern uint8_t pw2, pw3;

extern struct ve_at_t tune_me;
extern struct ve_stat_res_t res;

extern uint8_t tpsfuelcut;
uint8_t ego_warmup_count;

#ifdef RV8MT
uint8_t scantime, scantime_cnt;
#endif


volatile uint8_t alive;
//extern volatile uint8_t crank_enable_timeout;

/* to avoid conflict with LCD pins, LEDPORT is always set from userspace,
 * not interrupt 
 */
volatile uint8_t ledstate;
void setleds(void)
{
  LEDPORT = ((ledstate & LEDMASK 
       /*  LED_WARMUP  6 is masked, so that I see in the dark if voltage is too high
        * or low on the TACHIN pin
        */
#ifdef PINE2LED
        & ~64) | (PINE & 64)
#else
	)
#endif
        ) | (LEDPORT & ~LEDMASK);
}
/***************************************************************************/
/*** main entry point                                                    ***/
/***************************************************************************/
int main(void) {
#ifdef BUTTON
  uint8_t lcd_func=0;
#endif
  uint8_t lcd_update=4; // force update now
  uint8_t first_run=1;

  // init HW and misc. variables
  init();

  // schedule a priming pulse
  primepulse();

  // enable interrupts
  sei();


  // main loop
  for (;;) {
    setleds();
    calc_parameters();
    setleds();
    coolant_fan();
    setleds();
    fast_idle_valve();
    setleds();

    // only recalculate rpm if necessary
    if (engine.status_ext & _BV(new_rpm)) {
      cli();
      engine.status_ext &= ~_BV(new_rpm);
      sei();
      calc_rpm();
    }

    setleds();

    // make sure the engine is spinning
    if (engine.rpm) {
      // only enter cranking mode on power-on or if rpm drops to zero.
      if ((engine.status_ext & _BV(crank_enable)) && (engine.rpm <= config.cranking_thres)) {
	cranking();
      } else {    // running
	cli();
	engine.status_ext &= ~_BV(crank_enable);
	sei();


	// check if idle, and do idle control / anti-stall ?

        if (first_run) {  // this will force warmup enrichment
          first_run=0;
          cli();
          engine.status |= _BV(crank);
          sei();
        }
	warmup_enrich();

    setleds();
	tps_acc_enrich();
        closed_tps_decel();
	o2();
	ve_table_lookup();
    setleds();
	calc_total_enrichment();
      }
    }

    setleds();
    // every 1/10th second    
    if (rtc.flags & _BV(tenth_second)) {
      cli();
      rtc.flags &= ~_BV(tenth_second);
      sei();
	
      if (engine.status & _BV(running)) {
        if (!(config.iac[CONF] & _BV(disable_iac))) {
	  calc_iac_pos();
        }
	// Yeehaaaa, give me some VE values...
	if (config.ve_learn_conf & _BV(ve_learn_on))
	  ve_learn();
      }

      lcd_update++;
    }


    // every second
    if (rtc.flags & _BV(second)) {
      cli();
      rtc.flags &= ~_BV(second);
      sei();

      if (engine.status & _BV(running)) {  // engine is running,
	if (engine.status_ext & _BV(o2_not_ready)) { // and the O2 isn't ready
	  if (ego_warmup_count > 0) 
	    ego_warmup_count--; // wait some time
	  else
	    engine.status_ext &= ~_BV(o2_not_ready);  // and enable the O2
	}
      } else {
	if (ego_warmup_count < EGO_WARMUP)
	  ego_warmup_count++;
      }
    }

    setleds();
    store_control(); // store variables to eeprom

#ifdef BUTTON
#define MAX_LCD_MENUS 3 // this shouldn't be placed here
    if (rtc.flags & _BV(ten_ms)) {
      uint8_t but;
      cli();
      rtc.flags &= ~_BV(ten_ms);
      sei();

      but = button_read();
      if (but == 1) { // short press, show something different on the lcd
	// do some lcd magic
	lcd_func++;
	if (lcd_func > (MAX_LCD_MENUS-1))
	  lcd_func = 0;
      } else if (but == 2) { // long press
	save_variables();
	copyVEtoVEref();  // im case we are tuning the ve table
      }
    }
#endif

#ifdef LCD

    if (lcd_update >= 4) { // recalculate lcd data very 0.4 second
      lcd_update = 0;

    // print rpm and map
#ifdef LCD_4x20
    update_chars( "rpm:", 0);  // we can afford wasting a few chars in the 4x20 display
#else 
    lcd_setpos(0);
#endif

    int2BCDn(engine.rpm_hr, lcd_next_pos, BCD_DIGIT4);
    
    lcd_putstr("T");
    int2BCDn(sensors[MAT], lcd_next_pos, BCD_DIGIT3);
    int2BCDn(pwdisplay.bad, lcd_next_pos, BCD_DIGIT2);
    lcd_putstr("P"); // means MAP for LCD_4x20, Power for LCD_2x16
    
#ifdef LCD_4x20
    // print O2 and ego correction
    int2BCDn(engine.kpa, lcd_next_pos, BCD_DIGIT3);
    update_chars("O2 : 0.",  20);
#else
	if(pwdisplay.dirty){
		uint32_t divi;
		uint32_t power; // calculated from rpm and PW
		uint16_t rpm;	//  rpm, calced from TCNT3

/* After DIVDBY, time of AGGRNUM+1 periods will fit in this 16 bit:
 * 300 rpm = 5/sec => 10 events /sec, 100ms = 25000 (*4usec)
 *   *(AGGRNUM+1)  /  ....... < 65535
 */

// 250000 = 1sec
		divi = 60*250000 /INJ_PER_REV * (AGGRNUM+1)/(1<<ROTBY);
		rpm = divi / pwdisplay.persum ;

		power = rpm * (pwdisplay.pwsum - INJ_OPEN_CORR) ;
		int2BCD5(power>>8, lcd_next_pos); // 6000 * 2000 should fit
		int2BCD5(rpm, 16);	
		lcd_putstr(" W");
		int2BCD5( pwdisplay.pwsum, lcd_next_pos);
		pwdisplay.dirty = 0;
	}
   	{
   	char sp;

#ifdef KEYBOARD
/* TODO: called too often */
	if((sp = cooked_char())){
		if (sp == '9'){
			lcd_init();
			lcd_waitmode(1);
		}
	} else {
		sp=' ';
	}
#else
	sp=' ';
#endif
  	setleds();
	update_char(sp, 28);
	}
#endif
    {
      uint8_t t;
      t = 2*sensors[EGO];
      if (t > 99)
	t = 99;
      int2BCDn(t, lcd_next_pos, BCD_DIGIT2);
      lcd_putch('V');
    }

#ifdef LCD_4x20
    lcd_putstr(", ego: ");
    {
      uint8_t t;
      t = abs(corr.ego - 100);
      if (corr.ego > 100)
	lcd_putch('+');
      else
	lcd_putch('-');
      
      int2BCDn(t, lcd_next_pos, BCD_DIGIT2);
    }
/*
    // print engine temperature in the broken temperature scale
    update_chars("clt:  ",  40);
    int2BCDn(engine.coolant-40, lcd_next_pos, BCD_DIGIT3);
    lcd_putch('F');

    // print tps position
    lcd_putstr(", tps: ");
    int2BCDn(engine.tps, lcd_next_pos, BCD_DIGIT3);
*/
/*
update_chars("e",40);
int2BCD5(idle_error, lcd_next_pos);

update_chars(" i", lcd_next_pos);
int2BCD5(idle_integral, lcd_next_pos);

update_chars(" d", lcd_next_pos);
int2BCD5(idle_rpmdot, lcd_next_pos);
*/
/*
extern int16_t idle_integral;
extern int16_t idle_error;
extern int16_t rpmdot;
extern int16_t pos;
*/


#ifdef BUTTON

    if (lcd_func == 0) {
      // print IAC status
      update_chars("IAC: ", 60);
      int2BCD5(iac.position, lcd_next_pos);
      lcd_putstr(", ");
      if (iac.status & _BV(direction))
	lcd_putch('+');
      else
	lcd_putch('-');
      int2BCDn(iac.count, lcd_next_pos, BCD_DIGIT2);
      
      lcd_putstr("  ");
      int2BCDn(idle_rpm, lcd_next_pos, BCD_DIGIT3);

    } else if (lcd_func == 1) {
      // print VE learn status
      update_chars("VEL: r", 60);

      printhex(tune_me.rpm);
      lcd_putstr(" k");

      printhex(tune_me.kpa);
      lcd_putch(' ');

      if (tune_me.adjust < 0)
	lcd_putch('-');
      else
	lcd_putch('+');
      printhex(abs(tune_me.adjust));

      lcd_putstr(" c");
      printhex(tune_me.confidence);

    } else {
      lcd_setpos(60);
      lcd_putstr_P(PSTR("Are we having fun?  "));

    }

#else

    // print IAC information
    update_chars("pos: ", 60);
    int2BCD5(iac.position, lcd_next_pos);
    lcd_putstr(", ");
    if (iac.status & _BV(direction))
      lcd_putch('+');
    else
      lcd_putch('-');
    int2BCDn(iac.count, lcd_next_pos, BCD_DIGIT2);

    lcd_putstr("  ");
    int2BCDn(inj_port1.pw, lcd_next_pos, BCD_DIGIT3);
#endif // button

#endif // lcd

    }

    // finally refresh the lcd
    lcd_refresh();	// this writes maximum one char of the above writes!!!! The 
    // above should be split to several operations done in different cycles
#endif
    alive = 1;  // indicate that we've reached end of loop

#ifdef RV8MT
scantime = scantime_cnt;
scantime_cnt = 0;
#endif

  }

  return (0);
}



/***************************************************************************/
/*** init                                                                ***/
/***************************************************************************/
void init(void) {

  // init io
  PORTA = 0x0F; // misc LEDs
  DDRA = 0xF8;

  PORTB = 0x63; // pwm active low
  DDRB = 0xFC;

  PORTC = 0xE0; // stepmotor
  DDRC = 0x1F;

  PORTD = 0xFF;
  DDRD = 0x00;
  
  PORTE = 0xFF - 64; // we do NOT want pullup at the TACH in
  DDRE = 0x00;

  PORTF = 0x00; // adc channels
  DDRF = 0x00;

  PORTG = 0xFF;
  DDRG = 0x18;

#ifdef BUTTON
  // this isn't absolute necessary since the port is configured this way already,
  PORTE |= _BV(BUTTON_PIN);
  DDRE &= ~_BV(BUTTON_PIN);
  // input + pullup, button pulls the pin to ground
#endif

  if (PINB & _BV(J1))
    ms_version = 20;
  else
    ms_version = RV8_EMULATION;


#ifdef KEYBOARD
  keyboard_init();
#endif

#ifdef LCD
  lcd_init();
  //  update_chars("l1: -273C", 0);
  //  update_chars("l2: OK", 16); // changed 0 to O :-)
  lcd_waitmode(1);
  ledstate=0;
  setleds();
#endif

#ifdef MIK_LCD
	// used for LCD backlight power
  // fast pwm, clear on match - set on top, div256 prescaler == 244Hz
  TCCR2 = _BV(WGM21) | _BV(WGM20) | _BV(COM21) | _BV(CS22);
  OCR2 = LCD_BACKLIGHT_DUTYCYCLE;
#endif

#ifdef PWANAL
	eventqueue_init();
#endif

// enough to call once ;-)
#ifdef EVENTQUEUE
  eventqueue_init();
#endif

  // init timers
  // setup timer0 to interrupt at 10 kHz
  TCCR0 = _BV(WGM01) | _BV(CS01); // CTC, div8
  OCR0 = 200;
  TIMSK = _BV(OCIE0);  //output-compare interrupt

  // init pwm
  // Set OCnA/OCnB/OCnC on compare match, clear OCnA/OCnB/OCnC at TOP
  // fast pwm mode, TOP is stored in ICR1
  TCCR1A = _BV(WGM11);
  TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS10);
  ICR1 = PWM_FREQUENCY; // change pwm frequency

  // setup external interrupt
#ifdef PWANAL 
	EICRB = _BV(ISC60);  // falling or rising edge
#else
	EICRB = _BV(ISC61);  // falling edge
#endif

  EIMSK = _BV(INT6);  // enable interrupt

  // init rs232
  initUART();

  // load constants from eeprom to ram
  init_storage();
  load_variables();

  // init the sensor array
  initSensors();

  // get current barometric pressure
  sensors[BARO] = sensors[MAP];

  // init global variables
  calc_parameters();

  engine.status = 0;
  engine.status_ext = _BV(o2_not_ready) | _BV(crank_enable);
  engine.rpm = 0;
  engine.rpm_p = 0xFFFF; // just set it to some value that translates into a low rpm
  engine.rpm_c = 0;
  engine.rpm_hr = 0;
  engine.last_tps = engine.tps;

  inj_port1.pw = 0;
  inj_port2.pw = 0;

  ego_warmup_count = EGO_WARMUP;

  //  crank_enable_timeout = CRANK_TIMEOUT;

  init_fuelcalc();

  // init iac parameters
  init_iac();

  // enable fuelpump, 
  // it will be turned off within a couple of seconds, if engine isn't started
  PORTB |= _BV(FUELPUMP);
  // done, we have pressure, the fuelpump is really fast!


}


