/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen, mik@caffrey.dk]          *
***********************************************************
*  -- History --                                          *
*    release 0.1 - 06-03-2003                             *
*            http://caffrey.dk/megasquirt                 *
*    initial Motorola version by Bowling & Grippo (2002)  *
*            http://www.bgsoflex.com/megasquirt.html      *
\*********************************************************/


// THIS IS *VERY EXPERIMENTAL* ~ WORK IN PROGRESS
// DON'T USE IT UNLESS YOU WANT TO IMMOBILIZE (GUARANTEED) YOUR CAR!

/*
  ##### VE AUTO TUNING #####
  
  How it works:
    nobody knows yet...

    The function 've_learn' is called every 0.1 second. It stores [rpm,kpa,ego]
    in a buffer and when 'config.ve_learn_samples' is reached, the mean and 
    variance of all samples are calculated.

    The mean specifies which VE entry to modify and the variance gives a 
    figure of how much the data can be trusted. A small variance allows
    the VE entry to be changed much.

    The 'config.ve_learn_tau' is a timeconstant for changing the VE table. Even
    though the variance is 0, the VE table shouldn't change instantaniously.

  ##### NOTE #####

  There are still some debug leftovers in the code!

*/


#include "global.h"
#include "helpers.h"
#include "ve.h"
#include "fuelcalc.h"

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include <string.h>



volatile extern struct squirt_t inj_port1, inj_port2;
volatile extern struct engine_t engine;
volatile extern struct time_t rtc;
volatile extern uint8_t sensors[];
extern struct config_t config;
extern struct corr_t corr;
extern struct ve_t ve;

// these are global until we are out of the debug phase...
struct ve_at_t tune_me;
struct ve_stat_res_t res;


/***************************************************************************/
/*** VE table lookup                                                     ***/
/***************************************************************************/
void ve_table_lookup(void) {
  uint8_t ve_11, ve_12, ve_21, ve_22;
  uint8_t ve_low_kpa, ve_high_kpa;
  struct search_table_t kpa, rpm;

  if (config.config13 & _BV(CONTROL_STRATEGY)) { //Alpha-N
    engine.kpa = engine.tps;
  }

  search_table(ve.kpa_range, sizeof(ve.kpa_range), engine.kpa, &kpa);
  search_table(ve.rpm_range, sizeof(ve.rpm_range), engine.rpm, &rpm);

  // currently we ignore the fraction-part of the VE value
  ve_11 = ve.table[ VE_SIZE_RPM * (kpa.lindex) + (rpm.lindex)] >> 8;
  ve_12 = ve.table[ VE_SIZE_RPM * (kpa.lindex) +  rpm.uindex]    >> 8;
  ve_21 = ve.table[ VE_SIZE_RPM * kpa.uindex     + (rpm.lindex)] >> 8;
  ve_22 = ve.table[ VE_SIZE_RPM * kpa.uindex     +  rpm.uindex]    >> 8;

  ve_low_kpa = linear_interp(ve.rpm_range[rpm.lindex],
			     ve.rpm_range[rpm.uindex],
			     ve_11, ve_12, engine.rpm);

  ve_high_kpa = linear_interp(ve.rpm_range[rpm.lindex],
			      ve.rpm_range[rpm.uindex],
			      ve_21, ve_22, engine.rpm);

  corr.ve = linear_interp(ve.kpa_range[kpa.lindex],
			  ve.kpa_range[kpa.uindex],
			  ve_low_kpa, ve_high_kpa, engine.kpa);
}

/*
void ve_table_lookup(void) {
  uint8_t ve_11, ve_12, ve_21, ve_22;
  uint8_t ve_low_kpa, ve_high_kpa;
  struct search_table_t kpa, rpm;

  if (config.config13 & _BV(CONTROL_STRATEGY)) { //Alpha-N
    engine.kpa = engine.tps;
  }

  search_table(ve.kpa_range, sizeof(ve.kpa_range), engine.kpa, &kpa);
  search_table(ve.rpm_range, sizeof(ve.rpm_range), engine.rpm, &rpm);

  // currently we ignore the fraction-part of the VE value
  ve_11 = ve.table[ VE_SIZE_RPM * (kpa.index-1) + (rpm.index-1)] >> 8;
  ve_12 = ve.table[ VE_SIZE_RPM * (kpa.index-1) +  rpm.index]    >> 8;
  ve_21 = ve.table[ VE_SIZE_RPM * kpa.index     + (rpm.index-1)] >> 8;
  ve_22 = ve.table[ VE_SIZE_RPM * kpa.index     +  rpm.index]    >> 8;

  ve_low_kpa = linear_interp(rpm.lbound, rpm.ubound, ve_11, ve_12, engine.rpm);
  ve_high_kpa = linear_interp(rpm.lbound, rpm.ubound, ve_21, ve_22, engine.rpm);

  corr.ve = linear_interp(kpa.lbound, kpa.ubound, ve_low_kpa, ve_high_kpa, engine.kpa);
}
*/



/***************************************************************************/
/*** ve_entry_adjust                                                     ***/
/***************************************************************************/
/*
  ve_entry_adjust modifies a single VE entry pointed to by [rpm,kpa]. The 
  amount that the entry shall be changed is delivered in the adjust argument.

  ex: VE[i,j]=80, the new value is 83, so adjust will have the value +3.
*/
void ve_entry_adjust(uint8_t rpm_idx, uint8_t kpa_idx, uint8_t weight, int8_t adjust) {
  // if weight or adjust is zero, skip the rest
  if ((weight != 0) && (adjust != 0)) {  
    uint16_t ve_entry, ve_new, ve_candidate, limit;
    uint8_t ve_reference;

    // fetch current VE value and fraction
    ve_entry = ve.table[ VE_SIZE_RPM * kpa_idx + rpm_idx ];
    
    // calculate the candidate for the final VE value,
    // the lower 8 bits of ve_entry is the fraction
    ve_candidate = ve_entry + adjust * weight;

    // ve_new = ve_entry - ve_entry*weight/256 + ve_candidate*weight/256

    //    ve_new = (ve_entry * (256 - config.ve_learn_tau) + ve_candidate * config.ve_learn_tau) >> 8;

     // this has caused a lot of problems, the compiler chokes in the oneliner
    if (ve_candidate > ve_entry)
      ve_new = ve_entry + ((config.ve_learn_tau * (ve_candidate - ve_entry)) >> 8);
    else
      ve_new = ve_entry - ((config.ve_learn_tau * (ve_entry - ve_candidate)) >> 8);

    // limit the change, eg. compare ve_new with reference VE
    ve_reference = ve.ref_table[ VE_SIZE_RPM * kpa_idx + rpm_idx ];
    limit = (ve_reference * config.ve_learn_limit) / 100;

    if ((ve_new >> 8) > (ve_reference + limit)) {   // ve_new too large
      ve_new = ((ve_reference + limit) << 8) | 0xFF;
    } else if ((ve_new >> 8) < (ve_reference - limit)) {   // ve_new too small
      ve_new = (ve_reference - limit) << 8;
    }

    // store VE value and fraction
    ve.table[ VE_SIZE_RPM * kpa_idx + rpm_idx ] = ve_new;
  }
}



/***************************************************************************/
/*** ve_check_limit                                                      ***/
/***************************************************************************/
// check if rpm & kpa is within limits for safe modification of VE entry
// returns 1 if ok, 0 if power is too high
uint8_t ve_check_limit(uint8_t rpm, uint8_t kpa) {
  uint16_t power, max_power;

  power = rpm * kpa;
  max_power = (config.ve_learn_max_power[0] << 8) | config.ve_learn_max_power[1];

  if (power < max_power)
    return 1;
  else
    return 0;
}


/***************************************************************************/
/*** ve_calc_weights                                                     ***/
/***************************************************************************/
// calculate weights in the VE square
void ve_calc_weights(uint8_t rpm_idx, uint8_t kpa_idx, struct ve_at_t *a) {
  uint8_t drpm, dkpa;
  uint8_t dist_rpm, dist_kpa;
  uint8_t rpm_weight, kpa_weight;

  drpm = ve.rpm_range[rpm_idx] - ve.rpm_range[rpm_idx-1];
  dkpa = ve.kpa_range[kpa_idx] - ve.kpa_range[kpa_idx-1];

  // distance from VE-point to actual [rpm, kpa]
  dist_rpm = ve.rpm_range[rpm_idx] - a->rpm;
  dist_kpa = ve.kpa_range[kpa_idx] - a->kpa;
  
  // these are just for safety, dist_nnn must never be larger than dnnn
  if (dist_rpm > drpm)
    dist_rpm = drpm;

  if (dist_kpa > dkpa)
    dist_kpa = dkpa;

  // calculate weight
  if (drpm == 0)
    rpm_weight = 255;
  else
    // if the rpm bins are evenly spaced, the drpm division can be replaced by a fixed shift
    rpm_weight = 255 - (dist_rpm * 255) / drpm;

  if (dkpa == 0)
    kpa_weight = 255;
  else
    kpa_weight = 255 - (dist_kpa * 255) / dkpa;

  a->weight.rh_kh = (rpm_weight * kpa_weight) >> 8;
  a->weight.rh_kl = (rpm_weight * (255-kpa_weight)) >> 8;
  a->weight.rl_kh = ((255-rpm_weight) * kpa_weight) >> 8;
  a->weight.rl_kl = ((255-rpm_weight) * (255-kpa_weight)) >> 8;
}


/***************************************************************************/
/*** ve_auto_tune                                                        ***/
/***************************************************************************/
void ve_auto_tune(struct ve_at_t *a) {
  struct search_table_t kpa, rpm;

  // first find the 4 VE entries that should be evaluated
  search_table(ve.kpa_range, sizeof(ve.kpa_range), a->kpa, &kpa);
  search_table(ve.rpm_range, sizeof(ve.rpm_range), a->rpm, &rpm);

  // calculate weights
  ve_calc_weights(rpm.uindex, kpa.uindex, a);

  // adjust wrt confidence
  tune_me.adjust = (tune_me.adjust * tune_me.confidence) / 256;

  // then evaluate them
  ve_entry_adjust(rpm.lindex, kpa.lindex, a->weight.rl_kl, a->adjust);
  ve_entry_adjust(rpm.lindex, kpa.uindex, a->weight.rl_kh, a->adjust);
  ve_entry_adjust(rpm.uindex, kpa.lindex, a->weight.rh_kl, a->adjust);
  ve_entry_adjust(rpm.uindex, kpa.uindex, a->weight.rh_kh, a->adjust);

}
/*
void ve_auto_tune(struct ve_at_t *a) {
  struct search_table_t kpa, rpm;

  // first find the 4 VE entries that should be evaluated
  search_table(ve.kpa_range, sizeof(ve.kpa_range), a->kpa, &kpa);
  search_table(ve.rpm_range, sizeof(ve.rpm_range), a->rpm, &rpm);

  // calculate weights
  ve_calc_weights(rpm.index, kpa.index, a);

  // adjust wrt confidence
  tune_me.adjust = (tune_me.adjust * tune_me.confidence) / 256;

  // then evaluate them
  ve_entry_adjust(rpm.index-1, kpa.index-1, a->weight.rl_kl, a->adjust);
  ve_entry_adjust(rpm.index-1, kpa.index, a->weight.rl_kh, a->adjust);
  ve_entry_adjust(rpm.index, kpa.index-1, a->weight.rh_kl, a->adjust);
  ve_entry_adjust(rpm.index, kpa.index, a->weight.rh_kh, a->adjust);

}
*/

/***************************************************************************/
/*** ve_learn                                                            ***/
/***************************************************************************/
void ve_learn(void) {
  static uint8_t pfill;
  static uint8_t sample_count;
  static struct mfilter_t filter_rpm, filter_kpa, filter_ego;
  static struct ve_learn_stat_t learn_stat[STAT_SIZE];

  // store [rpm,kpa,ego] in circular buffer
  learn_stat[pfill].rpm = median_filter(&filter_rpm, engine.rpm);
  learn_stat[pfill].kpa = median_filter(&filter_kpa, engine.kpa);
  learn_stat[pfill].ego = median_filter(&filter_ego, corr.ego);
  pfill = (pfill + 1) % STAT_SIZE;

  // we don't want acceleration/deacceleration/low_coolant_temp to interfere
  if (!(((engine.status_ext & _BV(o2_closed_loop))) && (corr.warm == 100))) {

    // just clear the good_sample_count, the O2 correction is reset anyways when
    // tpsaen or tpsden are activated
    sample_count = 0;

  } else {

    // ok, still in closed loop, increase number of good samples
    sample_count++;
    
    if (sample_count >= VE_LEARN_SAMPLES+VE_LEARN_SKIP_SAMPLES) {
      //      struct ve_stat_res_t res;

      ve_calc_stat(learn_stat, STAT_SIZE, &res);
      
      if (ve_check_limit(res.rpm.mean, res.kpa.mean)) {
	uint8_t conf_rpm, conf_kpa, conf_ego;
	//      struct ve_at_t tune_me;

	conf_rpm = ve_variance_scale(res.rpm.variance, config.ve_learn_rpm_scale);
	conf_kpa = ve_variance_scale(res.kpa.variance, config.ve_learn_kpa_scale);
	conf_ego = ve_variance_scale(res.ego.variance, config.ve_learn_ego_scale);

	//	tune_me.cr = conf_rpm;
	//	tune_me.ck = conf_kpa;
	//	tune_me.ce = conf_ego;

	// the confidence depends on variation of all three variables
	tune_me.confidence = conf_rpm * ((conf_kpa * conf_ego) >> 8) >> 8;

	if (tune_me.confidence != 0) {
	  tune_me.rpm = res.rpm.mean;
	  tune_me.kpa = res.kpa.mean;
	  tune_me.adjust = (int16_t)res.ego.mean - 100;  // ego == 100 means no correction
	  ve_auto_tune(&tune_me);
	}
      }
      sample_count=0;
    }
  }
}


/***************************************************************************/
/*** ve_variance_scale                                                   ***/
/***************************************************************************/
uint8_t ve_variance_scale(uint8_t val, uint8_t scale) {
  uint16_t t = scale * val;
  
  if (t >= 255)
    return 0;
  else 
    return 255 - t;
}


/***************************************************************************/
/*** ve_calc_stat                                                        ***/
/***************************************************************************/
// calculate mean and variance of 'size' elements from buffer 't'
void ve_calc_stat(struct ve_learn_stat_t *t, uint8_t size, struct ve_stat_res_t *res) {
  uint8_t i;
  uint16_t mean_rpm=0, mean_kpa=0, mean_ego=0;
  uint16_t var_rpm=0, var_kpa=0, var_ego=0;

  // calculate mean
  for (i=0; i<size; i++) {
    mean_rpm += t[i].rpm;
    mean_kpa += t[i].kpa;
    mean_ego += t[i].ego;
  }
  
  mean_rpm = mean_rpm / size;
  mean_kpa = mean_kpa / size;
  mean_ego = mean_ego / size;

  // store the means
  res->rpm.mean = mean_rpm;
  res->kpa.mean = mean_kpa;
  res->ego.mean = mean_ego;

  // calculate variance
  for (i=0; i<size; i++) {
    int8_t drpm = t[i].rpm - mean_rpm;
    int8_t dkpa = t[i].kpa - mean_kpa;
    int8_t dego = t[i].ego - mean_ego;

    sum_square(drpm, &var_rpm);
    sum_square(dkpa, &var_kpa);
    sum_square(dego, &var_ego);
  }

  var_rpm = var_rpm / size;
  var_kpa = var_kpa / size;
  var_ego = var_ego / size;

  // store the variances
  res->rpm.variance = var_rpm;
  res->kpa.variance = var_kpa;
  res->ego.variance = var_ego;
}


/***************************************************************************/
/*** sum_square                                                          ***/
/***************************************************************************/
void sum_square(int8_t val, uint16_t *sum) {
  uint16_t t = *sum + val * val;

  if (t < *sum)
    *sum = 0xFFFF;  // overflow
  else 
    *sum = t;
}






/***************************************************************************/
/***************************************************************************/
/*** general purpose functions                                           ***/
/***************************************************************************/
/***************************************************************************/




/***************************************************************************/
/*** copyVEtoVEref                                                       ***/
/***************************************************************************/
void copyVEtoVEref(void) {
  uint8_t i;
  for (i=0; i<VE_SIZE_RPM*VE_SIZE_KPA; i++) {
    ve.ref_table[i] = ve.table[i] >> 8;
  }
}

/***************************************************************************/
/*** copyVEreftoVE                                                       ***/
/***************************************************************************/
void copyVEreftoVE(void) {
  uint8_t i;
  for (i=0; i<VE_SIZE_RPM*VE_SIZE_KPA; i++) {
    ve.table[i] = (ve.ref_table[i] << 8) | 0x7F;
  }
}


/***************************************************************************/
/*** interpolateVE                                                       ***/
/***************************************************************************/
// when using a 16x16 VE table, this function will fill the blanks if the 
// table is configured as a 8x8
void interpolateVE(void) {
  

}
