/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen, mik@caffrey.dk]          *
***********************************************************
*  -- History --                                          *
*    release 0.1 - 06-03-2003                             *
*            http://caffrey.dk/megasquirt                 *
*    initial Motorola version by Bowling & Grippo (2002)  *
*            http://www.bgsoflex.com/megasquirt.html      *
\*********************************************************/

#ifndef VE_H
#define VE_H

// length of buffer used for statistics
#define STAT_SIZE 8

// number of samples from the buffer to evaluate
#define VE_LEARN_SAMPLES 8

// upon _entry_ of closed loop, the might still be some acc enrichment
// skip the first n*0.1 seconds
#define VE_LEARN_SKIP_SAMPLES 10

// how many entries in the VE-table
#define VE_SIZE_RPM 8
#define VE_SIZE_KPA 8


struct ve_t {
  uint16_t table[VE_SIZE_RPM*VE_SIZE_KPA];     // main VE table (modified by VE learning) 
  uint8_t ref_table[VE_SIZE_RPM*VE_SIZE_KPA];  // reference VE table
  uint8_t rpm_range[VE_SIZE_RPM];  // VE table RPM bins for 2-D interpolation
  uint8_t kpa_range[VE_SIZE_KPA];  // VE table MAP bins for 2_D interpolation
};


struct ve_learn_stat_t {
  uint8_t rpm;
  uint8_t kpa;
  uint8_t ego;
};

struct mv_t {
  uint8_t mean;
  uint16_t variance;
};

struct ve_stat_res_t {
  struct mv_t rpm;
  struct mv_t kpa;
  struct mv_t ego;
};

struct ve_weight_t {  // box surrounding current [rpm, kpa]
  uint8_t rl_kl; // rpm low, kpa low
  uint8_t rl_kh; // rpm low, kpa high
  uint8_t rh_kl; // rpm high, kpa low
  uint8_t rh_kh; // rpm high, kpa high
};


struct ve_at_t {  // ve_auto_tune_t
  uint8_t rpm;
  uint8_t kpa;
  int8_t adjust;
  uint8_t confidence;
  //  uint8_t confidence1;
  struct ve_weight_t weight;
  //  uint8_t weight_rpm0_kpa0;
  //  uint8_t weight_rpm0_kpa1;
  //  uint8_t weight_rpm1_kpa0;
  //  uint8_t weight_rpm1_kpa1;
  //  uint8_t cr, ck, ce;
  //  uint8_t drpm, dkpa, dist_rpm, dist_kpa, rpm_weight, kpa_weight;
  //  uint8_t rpm_idx, kpa_idx;
  //  uint8_t rpm_idx1, kpa_idx1;
};


void ve_auto_tune(struct ve_at_t *a);
void ve_entry_adjust(uint8_t rpm_idx, uint8_t kpa_idx, uint8_t weight, int8_t adjust);
void ve_learn(void);
void ve_table_lookup(void);
uint8_t ve_check_limit(uint8_t rpm, uint8_t kpa);
void ve_calc_weights(uint8_t rpm_idx, uint8_t kpa_idx, struct ve_at_t *a);
void ve_calc_stat(struct ve_learn_stat_t *t, uint8_t size, struct ve_stat_res_t *res);
uint8_t ve_variance_scale(uint8_t val, uint8_t scale);

void sum_square(int8_t val, uint16_t *sum);
void copyVEtoVEref(void);
void copyVEreftoVE(void);


#endif
